#   ipmkchains.py -- Linux firewall rule manipulation library
#   Copyright (C) 2000  Bruce Guenter <bruceg@em.ca>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import random
import re
import string
import socket
import time

from ipmkchains_constants import *
from ipmkchains_types import *

_commands = [ ]
_commands_post = [ ]
_commands_pre = [ ]

_dst_defaults = { }

_dst_chains = { }

def command_pre(cmd):
    global _commands_pre
    _commands_pre.append(cmd)

def command(cmd):
    global _commands
    _commands.append(cmd)

def command_post(cmd):
    global _commands_post
    _commands_post.append(cmd)

command_pre('PATH=/sbin:/usr/sbin:/usr/local/sbin:$PATH')
command_pre('export PATH')

def commands():
    global _commands_pre
    global _commands
    global _commands_post
    return _commands_pre + _commands + _commands_post

def _tmpfile(base):
    while 1:
        filename = '/tmp/%s.tmp.%d.%f.%f' % (base, os.getpid(), time.time(),
                                             random.uniform(0, 1000000000))
        try:
            fd = os.open(filename, os.O_CREAT | os.O_EXCL, 0600)
            os.close(fd)
            break
        except OSError:
            continue
    return filename

def _read_fwchains(fwchains = '/proc/net/ip_fwchains',
                   fwnames = '/proc/net/ip_fwnames'):
    defaults = { }
    entries = { }
    for line in open(fwnames, 'r').readlines():
        (chain,default,rest) = string.split(line, ' ', 2)
        defaults[chain] = default
        entries[chain] = [ ]
    for line in open(fwchains, 'r').readlines():
        (chain,entry) = raw2rule(line)
        try:
            entries[chain].append(entry)
        except:
            entries[chain] = [entry]
    return defaults, entries

def flush_changes(lineno, srclen, inserts, deletes):
    while len(inserts) > 0 and len(deletes) > 0:
        (chain,rule) = raw2rule(inserts[0])
        cmd = 'ipchains -R %s %d %s' % (chain, lineno, rule.command())
        command(cmd)
        lineno = lineno + 1
        del inserts[0]
        del deletes[0]
    while len(inserts) > 0:
        (chain,rule) = raw2rule(inserts[0])
        cmd = rule.command()
        if srclen == 0 or lineno > srclen:
            cmd = 'ipchains -A %s %s' % (chain, cmd)
        elif lineno == 1:
            cmd = 'ipchains -I %s %s' % (chain, cmd)
        else:
            cmd = 'ipchains -I %s %d %s' % (chain, lineno, cmd)
        command(cmd)
        srclen = srclen + 1
        lineno = lineno + 1
        del inserts[0]
    while len(deletes) > 0:
        (chain,rule) = raw2rule(deletes[0])
        cmd = 'ipchains -D %s %d' % (chain, lineno)
        command(cmd)
        srclen = srclen - 1
        del deletes[0]
    return lineno, srclen

def _save_chain(name, chain, filename):
    f = open(filename, 'w')
    for i in range(len(chain)):
        f.write(chain[i].raw(name))
        f.write('\n')
    f.close()

def _diff_chains(chain, src, dst):
    tmpfile_src = _tmpfile('ipmkchains')
    tmpfile_dst = _tmpfile('ipmkchains')

    _save_chain(chain, src, tmpfile_src)
    _save_chain(chain, dst, tmpfile_dst)

    p = os.popen('diff -u -b %s %s' % (tmpfile_src, tmpfile_dst), 'r')
    result = map(string.rstrip, p.readlines())
    p.close()
    os.unlink(tmpfile_src)
    os.unlink(tmpfile_dst)
    return result

def _scan_chain(chain, src, dst, verbose):
    srclen = len(src)

    lineno = 0
    inserts = []
    deletes = []
    for line in _diff_chains(chain, src, dst):
        if verbose:
            command('# ' + line)
        if line[:4] == '--- ' or line[:4] == '+++ ':
            continue
        m = re.match(r'^@@ -(\d+),(\d+) \+(\d+),(\d+)', line)
        if m:
            (lineno,srclen) = flush_changes(lineno, srclen, inserts, deletes)
            lineno = int(m.group(3)) or 1
        elif line[0] == '-':
            deletes.append(line[1:])
        elif line[0] == '+':
            inserts.append(line[1:])
        elif line[0] == ' ':
            (lineno,srclen) = flush_changes(lineno, srclen, inserts, deletes)
            lineno = lineno + 1
    flush_changes(lineno, srclen, inserts, deletes)

def _set_default(chain, srcpolicy, dstpolicy):
    if srcpolicy <> dstpolicy:
        command('ipchains --policy %s %s' % (chain, dstpolicy))

def default(chain, target):
    if chain not in CHAINS:
        raise ValueError, "Can only set policy on built-in chains."
    global _dst_defaults
    _dst_defaults[chain] = target

_address_all = Address(IP_ALLZEROS,IP_ALLZEROS,0)

def rule(chain, target, proto=None,
         src=None, srcport=None, dst=None, dstport=None,
         iface=None, mark=0, output=0, tos=None, fragment=None,
         redirect=0, syn=None, log=None, bidirectional=None,
         icmptype=None):
    if (srcport or dstport) and proto not in [ICMP,TCP,UDP]:
        raise ValueError, "Can only specify ports for ICMP, TCP, or UDP"
    if redirect and target <> REDIRECT:
        raise ValueError, "Target must be REDIRECT to do redirection"
    
    if proto is None:
        proto = 0
    if src:
        src = Address(src)
    else:
        src = _address_all
    if dst:
        dst = Address(dst)
    else:
        dst = _address_all
    if not srcport:
        srcport1 = 0
        srcport2 = 65535
    elif type(srcport) is types.TupleType:
        srcport1 = srcport[0]
        srcport2 = srcport[1]
    else:
        srcport1 = srcport2 = srcport
    if not dstport:
        dstport1 = 0
        dstport2 = 65535
    elif type(dstport) is types.TupleType:
        dstport1 = dstport[0]
        dstport2 = dstport[1]
    else:
        dstport1 = dstport2 = dstport
    if not tos:
        tos = (0xff, 0x00)

    if icmptype:
        if proto <> ICMP:
            raise ValueError, "Protocol must by ICMP to use ICMP types."
        if icmp_types.has_key(icmptype):
            icmptype = icmp_types[icmptype]
        if type(icmptype) is not types.TupleType:
            raise ValueError, "Invalid ICMP type value '%s'" % icmptype
        if icmptype[0] >= 0:
            srcport1 = srcport2 = icmptype[0]
        else:
            srcport1 = 0
            srcport2 = 65535
        if icmptype[1] >= 0:
            dstport1 = dstport2 = icmptype[1]
        else:
            dstport1 = 0
            dstport2 = 65535

    flags2 = 0
    flags1 = 0
    if not iface:
        flags1 = flags1 | FLAG1_ANYIFACE
        iface = '-'
    if log:
        flags1 = flags1 | FLAG1_LOG
    if syn:
        flags1 = flags1 | FLAG1_SYN
        if syn < 0:
            flags2 = flags2 | FLAG2_SYN
    if fragment:
        flags1 = flags1 | FLAG1_FRAG
        if fragment < 0:
            flags2 = flags2 | FLAG2_FRAG
    if mark:
        flags1 = flags1 | FLAG1_SETMARK
    if output:
        flags1 = flags1 | FLAG1_OUTPUT
    if src.negate:
        flags2 = flags2 | FLAG2_SRCIP
    if dst.negate:
        flags2 = flags2 | FLAG2_DSTIP
    if srcport1 < 0:
        srcport1 = -srcport1
        flags2 = flags2 | FLAG2_SRCPORT
    if dstport1 < 0:
        dstport1 = -dstport1
        flags2 = flags2 | FLAG2_DSTPORT
    if srcport2 < 0:
        srcport2 = -srcport2
        flags2 = flags2 | FLAG2_SRCPORT
    if dstport2 < 0:
        dstport2 = -dstport2
        flags2 = flags2 | FLAG2_DSTPORT
    if iface[0] == '-' and len(iface) > 1:
        iface = iface[1:]
        flags2 = flags2 | FLAG2_IFACE

    r = Rule(target=target, proto=proto, flags1=flags1, flags2=flags2,
             tosa=tos[0], tosx=tos[1], iface=iface, mark=mark,
             maxsize=output, redirect=redirect,
             srcip=src.ip, srcmask=src.mask,
             srcport1=srcport1, srcport2=srcport2,
             dstip=dst.ip, dstmask=dst.mask,
             dstport1=dstport1, dstport2=dstport2)
    global _dst_chains
    try:
        _dst_chains[chain].append(r)
    except:
        _dst_chains[chain] = [r]
    if bidirectional:
        r = Rule(target=target, proto=proto, flags1=flags1, flags2=flags2,
                 tosa=tos[0], tosx=tos[1], iface=iface, mark=mark,
                 maxsize=output, redirect=redirect,
                 srcip=dst.ip, srcmask=dst.mask,
                 srcport1=dstport1, srcport2=dstport2,
                 dstip=src.ip, dstmask=src.mask,
                 dstport1=srcport1, dstport2=srcport2)
        _dst_chains[chain].append(r)

def add_route_rules(chain,prefix='chk_',log=1):
    f = open('/proc/net/route', 'r')
    f.readline()
    routes = []
    interfaces = {}
    # Collect the list of routes and interfaces from the routing table
    for line in f.readlines():
        route = raw2route(line)
        if route.destination.ip == IP_ALLZEROS:
	    default_interface = route.interface
	    if interfaces.has_key(route.interface):
	        del interfaces[route.interface]
	elif route.destination.mask <> IP_ALLONES:
            routes.append(route)
            interfaces[route.interface] = 1
    interfaces = interfaces.keys()
    interfaces.sort()
    rule(chain, prefix+'dflt', iface=default_interface)
    for interface in interfaces:
        rule(chain, prefix+interface, iface=interface)
    for route in routes:
	if route.interface <> default_interface:
            rule(prefix+route.interface, RETURN, src=route.destination)
            rule(prefix+'dflt', DENY, src=route.destination, log=log)
    for interface in interfaces:
        rule(prefix+interface, DENY, log=log)

def install(test=0, verbose=0):
    src_defaults, src_chains = _read_fwchains()
    global _dst_chains
    global _dst_defaults
    for chain in src_chains.keys():
        if not _dst_chains.has_key(chain) and chain not in CHAINS:
            command_post('ipchains --flush %s' % chain)
            command_post('ipchains --delete-chain %s' % chain)
    for chain in _dst_chains.keys():
        if chain in CHAINS:
            continue
        if not src_chains.has_key(chain):
            src_chains[chain] = []
            command_pre('ipchains --new %s' % chain)
        _scan_chain(chain, src_chains[chain], _dst_chains[chain], verbose)
    for chain in CHAINS:
        # Set default before scanning chain
        _set_default(chain, src_defaults.get(chain, DENY),
                     _dst_defaults.get(chain, DENY))
        if _dst_chains.has_key(chain):
            _scan_chain(chain, src_chains[chain], _dst_chains[chain], verbose)
    list = string.join(commands(), '\n')
    if test:
        print list
    else:
        os.system(list)
