#   ipmkchains_types.py -- Types for ipmkchains
#   Copyright (C) 2000  Bruce Guenter <bruceg@em.ca>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import re
import string
import types

from ipmkchains_constants import *

_RX_DOTQ = re.compile(r'^(\d+)\.(\d+)\.(\d+)\.(\d+)$')

def hexint(str):
    return string.atoi(str, 16)

def hex2ip(str):
    return (string.atoi(str[0:2], 16),
            string.atoi(str[2:4], 16),
            string.atoi(str[4:6], 16),
            string.atoi(str[6:8], 16))

def rhex2ip(str):
    return (string.atoi(str[6:8], 16),
            string.atoi(str[4:6], 16),
            string.atoi(str[2:4], 16),
            string.atoi(str[0:2], 16))

def ip2hex(ip):
    return '%02x%02x%02x%02x' % ip

def mask2ip(int):
    bits = 0xffffffff << (32-int)
    return ((bits >> 24) & 0xff,
            (bits >> 16) & 0xff,
            (bits >> 8) & 0xff,
            bits & 0xff)

def auto2ip(data):
    td = type(data)
    if td == types.TupleType:
        return data
    if td == types.ListType:
        return tuple(data)
    if td == types.StringType:
        return tuple(map(int, _RX_DOTQ.match(data).groups()))
    if td == types.IntType:
        return mask2ip(data)

def ip2str(ip):
    return '%d.%d.%d.%d' % ip

###############################################################################

_RX_ADDR = re.compile(r'^(.+)/(.+)->(.+)/(.+)$')
_RX_PORTS = re.compile(r'^(\d+)-(\d+)$')
_RX_WHITESPACE = re.compile(r'\s+')

_rule_defaults = {
    'target': '-',
    'proto': 0,
    'flags1': 0x08,
    'flags2': 0x00,
    'tosx': 0x00,
    'tosa': 0xff,
    'iface': '-',
    'mark': 0,
    'maxsize': 0,
    'srcip': (0,0,0,0),
    'srcmask': (0,0,0,0),
    'srcport1': 0,
    'srcport2': 65535,
    'dstip': (0,0,0,0),
    'dstmask': (0,0,0,0),
    'dstport1': 0,
    'dstport2': 65535,
    'redirect': 0,
    }

class Rule:
    def __init__(self, **values):
        for (key,value) in _rule_defaults.items():
            setattr(self, key, value)
        for (key,value) in values.items():
            if not _rule_defaults.has_key(key):
                raise AttributeError, "'%s' is an invalid value name"%key
            setattr(self, key, value)
        self.srcip = auto2ip(self.srcip)
        self.srcmask = auto2ip(self.srcmask)
        self.dstip = auto2ip(self.dstip)
        self.dstmask = auto2ip(self.dstmask)
        
    def __cmp__(self, that):
        for (key,val) in self.__dict__.items():
            c = cmp(val, getattr(that, key))
            if c:
                return c
        return 0

    def cmdpart(self, name, negflag, value=''):
        if self.flags2 & negflag:
            if value:
                return ' --%(name)s ! %(value)s' % vars()
            else:
                return ' ! --%s' % name
        else:
            return ' --%(name)s %(value)s' % vars()

        
    def command(self):
        s = '--jump %s' % self.target
        if self.proto:
            if self.redirect and self.target == REDIRECT:
                s = s + " %d" % self.redirect
            s = s + self.cmdpart('proto', FLAG2_PROTO, self.proto)
        if not (self.flags1 & FLAG1_ANYIFACE):
            s = s + self.cmdpart('interface', FLAG2_IFACE, self.iface)
        if self.flags1 & FLAG1_SETMARK or self.mark < 0:
            s = s + ' --mark %d' % self.mark
        elif self.mark:
            s = s + ' --mark +%d' % self.mark
        if self.srcip <> (0,0,0,0) or self.srcmask <> (0,0,0,0):
            s = s + self.cmdpart('source', FLAG2_SRCIP,
                                 '%d.%d.%d.%d/%d.%d.%d.%d'%(
                self.srcip+self.srcmask))
        if self.dstip <> (0,0,0,0) or self.dstmask <> (0,0,0,0):
            s = s + self.cmdpart('destination', FLAG2_DSTIP,
                                 '%d.%d.%d.%d/%d.%d.%d.%d'%(
                self.dstip+self.dstmask))
        if self.proto:
            if self.srcport1 <> _rule_defaults['srcport1'] or \
               self.srcport2 <> _rule_defaults['srcport2']:
                s = s + self.cmdpart('source-port', FLAG2_SRCPORT,
                                     '%d:%d' % (self.srcport1, self.srcport2))
            if self.dstport1 <> _rule_defaults['dstport1'] or \
               self.dstport2 <> _rule_defaults['dstport2']:
                s = s + self.cmdpart('destination-port', FLAG2_DSTPORT,
                                     '%d:%d' % (self.dstport1, self.dstport2))
        if self.flags1 & FLAG1_LOG:
            s = s + ' --log'
        if self.flags1 & FLAG1_SYN:
            s = s + self.cmdpart('syn', FLAG2_SYN)
        if self.flags1 & FLAG1_FRAG:
            s = s + self.cmdpart('fragment', FLAG2_FRAG)
        if self.flags1 & FLAG1_OUTPUT:
            s = s + ' --output %d' % self.maxsize
        if self.tosa <> _rule_defaults['tosa'] or \
           self.tosx <> _rule_defaults['tosx']:
            s = s + ' --TOS %x %x' % (self.tosa, self.tosx)
        return s

    def raw(self, chain):
        return ("%s %s/%s->%s/%s %s %x %x %d 0 0 0 0 "
                " %d-%d %d-%d A%x X%x %08x %d %d %s" % (
            chain, ip2hex(self.srcip), ip2hex(self.srcmask),
            ip2hex(self.dstip), ip2hex(self.dstmask),
            self.iface, self.flags1, self.flags2, self.proto,
            self.srcport1, self.srcport2, self.dstport1, self.dstport2,
            self.tosa, self.tosx, self.redirect, self.mark,
            self.maxsize, self.target ) )

def raw2rule(line):
    line = string.strip(line)
    (chain,
     addrs,iface,flags1,flags2,proto,phi,plo,bhi,blo,
     srcports,dstports,tosa,tosx,
     redirect,mark,maxsize,target) = _RX_WHITESPACE.split(line)
    (srcip,srcmask,
     dstip,dstmask) = map(hex2ip, _RX_ADDR.match(addrs).groups())
    (srcport1,srcport2) = map(int, _RX_PORTS.match(srcports).groups())
    (dstport1,dstport2) = map(int, _RX_PORTS.match(dstports).groups())
    
    flags1 = hexint(flags1)
    flags2 = hexint(flags2)
    proto = int(proto)
    tosa = hexint(tosa[1:])
    tosx = hexint(tosx[1:])
    redirect = hexint(redirect)
    mark = int(mark)
    maxsize = int(maxsize)
    return (chain,
            Rule(target=target, proto=proto,
                 flags1=flags1, flags2=flags2,
                 tosa=tosa, tosx=tosx, iface=iface,
                 mark=mark, maxsize=maxsize,
                 srcip=srcip, srcmask=srcmask,
                 srcport1=srcport1, srcport2=srcport2,
                 dstip=dstip, dstmask=dstmask,
                 dstport1=dstport1, dstport2=dstport2,
                 redirect=redirect )
            )
###############################################################################

_RX_MASKEDIP = re.compile(r'^(\d+\.\d+\.\d+\.\d+)/(\d+)$')

class Address:
    def __init__(self, ip, mask=None, negate=None):
        # If the first arg is the same as this, use it to set the values of
        # IP, mask, and negate only if they aren't also set.
        if type(ip) is type(self):
            if mask is None:
                mask = ip.mask
            if negate is None:
                negate = ip.negate
            ip = ip.ip
        # Set defaults for network mask and negate flag
        if mask is None:
            mask = IP_ALLONES
        if negate is None:
            negate = 0
        # Convert strings of the form "a.b.c.d/m" to IP/mask
        if type(ip) is types.StringType:
            m = _RX_MASKEDIP.match(ip)
            if m:
                ip = auto2ip(m.group(1))
                mask = int(m.group(2))
        self.ip = auto2ip(ip)
        self.mask = auto2ip(mask)
        self.negate = negate
    def __neg__(self):
        return Address(self, negate=(not self.negate))
    def __repr__(self):
        s = 'Address(' + str(self.ip)
        if self.mask <> IP_ALLONES:
            s = '%s, %s' % (s, self.mask)
        if self.negate:
            s = '%s, negate=%d' % (s, self.negate)
        return s + ')'
    def __str__(self):
        s = ''
        if self.negate: s = '!'
        s = s + string.join(map(str, self.ip), '.')
        if self.mask <> IP_ALLONES:
            try:
                mask = str(mask2bits[self.mask])
            except:
                mask = string.join(map(str, self.mask), '.')
            s = '%s/%s' % (s, mask)
        return s

# The following is included only for backwards compatibility
address = Address

###############################################################################
class Route:
    def __init__(self, dest, gateway, interface):
        self.destination = dest
        self.gateway = gateway
        self.interface = interface
    def __repr__(self):
        return "Route(%s, %s, '%s')" % (repr(self.destination),
                                        repr(self.gateway),
                                        self.interface)
    def __str__(self):
        return '%s -> %s (%s)' % (self.destination,
                                  self.gateway,
                                  self.interface)

def raw2route(line):
    line = string.strip(line)
    (interface,destination,gateway,flags,refcnt,use,metric,mask,
     mtu,window,irtt) = _RX_WHITESPACE.split(line)
    return Route(Address(rhex2ip(destination), rhex2ip(mask)),
                 Address(rhex2ip(gateway)), interface)
